from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, Literal, TypeVar, cast

from pystac import (
    Asset,
    Catalog,
    Collection,
    Item,
    ItemAssetDefinition,
    Link,
    STACError,
)
from pystac.extensions.classification import ClassificationExtension
from pystac.extensions.datacube import DatacubeExtension
from pystac.extensions.eo import EOExtension
from pystac.extensions.file import FileExtension
from pystac.extensions.grid import GridExtension
from pystac.extensions.item_assets import ItemAssetsExtension
from pystac.extensions.mgrs import MgrsExtension
from pystac.extensions.mlm import (
    AssetDetailedMLMExtension,
    AssetGeneralMLMExtension,
    MLMExtension,
)
from pystac.extensions.pointcloud import PointcloudExtension
from pystac.extensions.projection import ProjectionExtension
from pystac.extensions.raster import RasterExtension
from pystac.extensions.render import Render, RenderExtension
from pystac.extensions.sar import SarExtension
from pystac.extensions.sat import SatExtension
from pystac.extensions.scientific import ScientificExtension
from pystac.extensions.storage import StorageExtension
from pystac.extensions.table import TableExtension
from pystac.extensions.timestamps import TimestampsExtension
from pystac.extensions.version import BaseVersionExtension, VersionExtension
from pystac.extensions.view import ViewExtension
from pystac.extensions.xarray_assets import XarrayAssetsExtension

#: Generalized version of :class:`~pystac.Asset`,
#: :class:`~pystac.ItemAssetDefinition`, or :class:`~pystac.Link`
T = TypeVar("T", Asset, ItemAssetDefinition, Link)
#: Generalized version of :class:`~pystac.Asset` or
#: :class:`~pystac.ItemAssetDefinition`
U = TypeVar("U", Asset, ItemAssetDefinition)

EXTENSION_NAMES = Literal[
    "classification",
    "cube",
    "eo",
    "file",
    "grid",
    "item_assets",
    "mgrs",
    "mlm",
    "pc",
    "proj",
    "raster",
    "render",
    "sar",
    "sat",
    "sci",
    "storage",
    "table",
    "timestamps",
    "version",
    "view",
    "xarray",
]

EXTENSION_NAME_MAPPING: dict[EXTENSION_NAMES, Any] = {
    ClassificationExtension.name: ClassificationExtension,
    DatacubeExtension.name: DatacubeExtension,
    EOExtension.name: EOExtension,
    FileExtension.name: FileExtension,
    GridExtension.name: GridExtension,
    ItemAssetsExtension.name: ItemAssetsExtension,
    MgrsExtension.name: MgrsExtension,
    MLMExtension.name: MLMExtension,
    PointcloudExtension.name: PointcloudExtension,
    ProjectionExtension.name: ProjectionExtension,
    RasterExtension.name: RasterExtension,
    RenderExtension.name: RenderExtension,
    SarExtension.name: SarExtension,
    SatExtension.name: SatExtension,
    ScientificExtension.name: ScientificExtension,
    StorageExtension.name: StorageExtension,
    TableExtension.name: TableExtension,
    TimestampsExtension.name: TimestampsExtension,
    VersionExtension.name: VersionExtension,
    ViewExtension.name: ViewExtension,
    XarrayAssetsExtension.name: XarrayAssetsExtension,
}


def _get_class_by_name(name: str) -> Any:
    try:
        return EXTENSION_NAME_MAPPING[cast(EXTENSION_NAMES, name)]
    except KeyError as e:
        raise KeyError(
            f"Extension '{name}' is not a valid extension. "
            f"Options are {list(EXTENSION_NAME_MAPPING)}"
        ) from e


@dataclass
class CatalogExt:
    """Supporting the :attr:`~pystac.Catalog.ext` accessor for interacting
    with extension classes
    """

    stac_object: Catalog

    def has(self, name: EXTENSION_NAMES) -> bool:
        """Whether the given extension is enabled on this STAC object

        Args:
            name : Extension identifier (eg: 'eo')

        Returns:
            bool: ``True`` if extension is enabled, otherwise ``False``
        """
        return cast(bool, _get_class_by_name(name).has_extension(self.stac_object))

    def add(self, name: EXTENSION_NAMES) -> None:
        """Add the given extension to this STAC object

        Args:
            name : Extension identifier (eg: 'eo')
        """
        _get_class_by_name(name).add_to(self.stac_object)

    def remove(self, name: EXTENSION_NAMES) -> None:
        """Remove the given extension from this STAC object

        Args:
            name : Extension identifier (eg: 'eo')
        """
        _get_class_by_name(name).remove_from(self.stac_object)

    @property
    def version(self) -> VersionExtension[Catalog]:
        return VersionExtension.ext(self.stac_object)


@dataclass
class CollectionExt(CatalogExt):
    """Supporting the :attr:`~pystac.Collection.ext` accessor for interacting
    with extension classes
    """

    stac_object: Collection

    @property
    def cube(self) -> DatacubeExtension[Collection]:
        return DatacubeExtension.ext(self.stac_object)

    @property
    def item_assets(self) -> dict[str, ItemAssetDefinition]:
        return ItemAssetsExtension.ext(self.stac_object).item_assets

    @property
    def mlm(self) -> MLMExtension[Collection]:
        return MLMExtension.ext(self.stac_object)

    @property
    def render(self) -> dict[str, Render]:
        return RenderExtension.ext(self.stac_object).renders

    @property
    def sci(self) -> ScientificExtension[Collection]:
        return ScientificExtension.ext(self.stac_object)

    @property
    def table(self) -> TableExtension[Collection]:
        return TableExtension.ext(self.stac_object)

    @property
    def xarray(self) -> XarrayAssetsExtension[Collection]:
        return XarrayAssetsExtension.ext(self.stac_object)


@dataclass
class ItemExt:
    """Supporting the :attr:`~pystac.Item.ext` accessor for interacting
    with extension classes
    """

    stac_object: Item

    def has(self, name: EXTENSION_NAMES) -> bool:
        """Whether the given extension is enabled on this STAC object

        Args:
            name : Extension identifier (eg: 'eo')

        Returns:
            bool: ``True`` if extension is enabled, otherwise ``False``
        """
        return cast(bool, _get_class_by_name(name).has_extension(self.stac_object))

    def add(self, name: EXTENSION_NAMES) -> None:
        """Add the given extension to this STAC object

        Args:
            name : Extension identifier (eg: 'eo')
        """
        _get_class_by_name(name).add_to(self.stac_object)

    def remove(self, name: EXTENSION_NAMES) -> None:
        """Remove the given extension from this STAC object

        Args:
            name : Extension identifier (eg: 'eo')
        """
        _get_class_by_name(name).remove_from(self.stac_object)

    @property
    def classification(self) -> ClassificationExtension[Item]:
        return ClassificationExtension.ext(self.stac_object)

    @property
    def cube(self) -> DatacubeExtension[Item]:
        return DatacubeExtension.ext(self.stac_object)

    @property
    def eo(self) -> EOExtension[Item]:
        return EOExtension.ext(self.stac_object)

    @property
    def grid(self) -> GridExtension:
        return GridExtension.ext(self.stac_object)

    @property
    def mgrs(self) -> MgrsExtension:
        return MgrsExtension.ext(self.stac_object)

    @property
    def mlm(self) -> MLMExtension[Item]:
        return MLMExtension.ext(self.stac_object)

    @property
    def pc(self) -> PointcloudExtension[Item]:
        return PointcloudExtension.ext(self.stac_object)

    @property
    def proj(self) -> ProjectionExtension[Item]:
        return ProjectionExtension.ext(self.stac_object)

    @property
    def render(self) -> RenderExtension[Item]:
        return RenderExtension.ext(self.stac_object)

    @property
    def sar(self) -> SarExtension[Item]:
        return SarExtension.ext(self.stac_object)

    @property
    def sat(self) -> SatExtension[Item]:
        return SatExtension.ext(self.stac_object)

    @property
    def sci(self) -> ScientificExtension[Item]:
        return ScientificExtension.ext(self.stac_object)

    @property
    def storage(self) -> StorageExtension[Item]:
        return StorageExtension.ext(self.stac_object)

    @property
    def table(self) -> TableExtension[Item]:
        return TableExtension.ext(self.stac_object)

    @property
    def timestamps(self) -> TimestampsExtension[Item]:
        return TimestampsExtension.ext(self.stac_object)

    @property
    def version(self) -> VersionExtension[Item]:
        return VersionExtension.ext(self.stac_object)

    @property
    def view(self) -> ViewExtension[Item]:
        return ViewExtension.ext(self.stac_object)

    @property
    def xarray(self) -> XarrayAssetsExtension[Item]:
        return XarrayAssetsExtension.ext(self.stac_object)


class _AssetsExt(Generic[T]):
    stac_object: T

    def has(self, name: EXTENSION_NAMES) -> bool:
        """Whether the given extension is enabled on the owner

        Args:
            name : Extension identifier (eg: 'eo')

        Returns:
            bool: ``True`` if extension is enabled, otherwise ``False``
        """
        if self.stac_object.owner is None:
            raise STACError(
                f"Attempted to use `.ext.has('{name}') for an object with no owner. "
                "Use `.set_owner` and then try to check the extension again."
            )
        else:
            return cast(
                bool, _get_class_by_name(name).has_extension(self.stac_object.owner)
            )

    def add(self, name: EXTENSION_NAMES) -> None:
        """Add the given extension to the owner

        Args:
            name : Extension identifier (eg: 'eo')
        """
        if self.stac_object.owner is None:
            raise STACError(
                f"Attempted to add extension='{name}' for an object with no owner. "
                "Use `.set_owner` and then try to add the extension again."
            )
        else:
            _get_class_by_name(name).add_to(self.stac_object.owner)

    def remove(self, name: EXTENSION_NAMES) -> None:
        """Remove the given extension from the owner

        Args:
            name : Extension identifier (eg: 'eo')
        """
        if self.stac_object.owner is None:
            raise STACError(
                f"Attempted to remove extension='{name}' for an object with no owner. "
                "Use `.set_owner` and then try to remove the extension again."
            )
        else:
            _get_class_by_name(name).remove_from(self.stac_object.owner)


class _AssetExt(_AssetsExt[U]):
    stac_object: U

    @property
    def classification(self) -> ClassificationExtension[U]:
        return ClassificationExtension.ext(self.stac_object)

    @property
    def cube(self) -> DatacubeExtension[U]:
        return DatacubeExtension.ext(self.stac_object)

    @property
    def eo(self) -> EOExtension[U]:
        return EOExtension.ext(self.stac_object)

    @property
    def pc(self) -> PointcloudExtension[U]:
        return PointcloudExtension.ext(self.stac_object)

    @property
    def proj(self) -> ProjectionExtension[U]:
        return ProjectionExtension.ext(self.stac_object)

    @property
    def raster(self) -> RasterExtension[U]:
        return RasterExtension.ext(self.stac_object)

    @property
    def sar(self) -> SarExtension[U]:
        return SarExtension.ext(self.stac_object)

    @property
    def sat(self) -> SatExtension[U]:
        return SatExtension.ext(self.stac_object)

    @property
    def storage(self) -> StorageExtension[U]:
        return StorageExtension.ext(self.stac_object)

    @property
    def table(self) -> TableExtension[U]:
        return TableExtension.ext(self.stac_object)

    @property
    def version(self) -> BaseVersionExtension[U]:
        return BaseVersionExtension.ext(self.stac_object)

    @property
    def view(self) -> ViewExtension[U]:
        return ViewExtension.ext(self.stac_object)


@dataclass
class AssetExt(_AssetExt[Asset]):
    """Supporting the :attr:`~pystac.Asset.ext` accessor for interacting
    with extension classes
    """

    stac_object: Asset

    @property
    def file(self) -> FileExtension[Asset]:
        return FileExtension.ext(self.stac_object)

    @property
    def mlm(self) -> AssetGeneralMLMExtension[Asset] | AssetDetailedMLMExtension:
        if "mlm:name" in self.stac_object.extra_fields:
            return AssetDetailedMLMExtension.ext(self.stac_object)
        else:
            return AssetGeneralMLMExtension.ext(self.stac_object)

    @property
    def timestamps(self) -> TimestampsExtension[Asset]:
        return TimestampsExtension.ext(self.stac_object)

    @property
    def xarray(self) -> XarrayAssetsExtension[Asset]:
        return XarrayAssetsExtension.ext(self.stac_object)


@dataclass
class ItemAssetExt(_AssetExt[ItemAssetDefinition]):
    """Supporting the :attr:`~pystac.ItemAssetDefinition.ext` accessor for interacting
    with extension classes
    """

    stac_object: ItemAssetDefinition

    @property
    def mlm(self) -> MLMExtension[ItemAssetDefinition]:
        return MLMExtension.ext(self.stac_object)


@dataclass
class LinkExt(_AssetsExt[Link]):
    """Supporting the :attr:`~pystac.Link.ext` accessor for interacting
    with extension classes
    """

    stac_object: Link

    @property
    def file(self) -> FileExtension[Link]:
        return FileExtension.ext(self.stac_object)



from __future__ import annotations

import json
import os
from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import pystac
from pystac.serialization import (
    identify_stac_object,
    identify_stac_object_type,
    merge_common_properties,
    migrate_to_latest,
)
from pystac.utils import HREF, _is_url, safe_urlparse

# Use orjson if available
try:
    import orjson
except ImportError:
    orjson = None  # type: ignore[assignment]

# Is urllib3 available?
try:
    import urllib3  # noqa
except ImportError:
    HAS_URLLIB3 = False
else:
    HAS_URLLIB3 = True

if TYPE_CHECKING:
    from pystac.catalog import Catalog
    from pystac.stac_object import STACObject


class StacIO(ABC):
    _default_io: Callable[[], StacIO] | None = None

    def __init__(self, headers: dict[str, str] | None = None):
        self.headers = headers or {}

    @abstractmethod
    def read_text(self, source: HREF, *args: Any, **kwargs: Any) -> str:
        """Read text from the given URI.

        The source to read from can be specified as a string or :class:`os.PathLike`
        object (:class:`~pystac.Link` is a path-like object). If it is a string, it
        must be a URI or local path from which to read. Using a :class:`~pystac.Link`
        enables implementations to use additional link information, such as paging
        information contained in the extended links described in the `STAC API spec
        <https://github.com/radiantearth/stac-api-spec/tree/master/item-search#paging>`__.

        Args:
            source : The source to read from.
            *args : Arbitrary positional arguments that may be utilized by the concrete
                implementation.
            **kwargs : Arbitrary keyword arguments that may be utilized by the concrete
                implementation.

        Returns:
            str: The text contained in the file at the location specified by the uri.
        """
        raise NotImplementedError

    @abstractmethod
    def write_text(
        self,
        dest: HREF,
        txt: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Write the given text to a file at the given URI.

        The destination to write to can be specified as a string or
        :class:`os.PathLike` object (:class:`~pystac.Link` is a path-like object). If
        it is a string, it must be a URI or local path from which to read. Using a
        :class:`~pystac.Link` enables implementations to use additional link
        information.

        Args:
            dest : The destination to write to.
            txt : The text to write.
        """
        raise NotImplementedError

    def json_loads(self, txt: str, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Method used internally by :class:`StacIO` instances to deserialize a
        dictionary from a JSON string.

        This method may be overwritten in :class:`StacIO` subclasses to provide custom
        deserialization logic. The method accepts arbitrary keyword arguments. These are
        not used by the default implementation, but may be used by subclass
        implementations.

        Args:

            txt : The JSON string to deserialize to a dictionary.
        """
        result: dict[str, Any]
        if orjson is not None:
            result = orjson.loads(txt)
        else:
            result = json.loads(txt, *args, **kwargs)
        return result

    def json_dumps(self, json_dict: dict[str, Any], *args: Any, **kwargs: Any) -> str:
        """Method used internally by :class:`StacIO` instances to serialize a dictionary
        to a JSON string.

        This method may be overwritten in :class:`StacIO` subclasses to provide custom
        serialization logic. The method accepts arbitrary keyword arguments. These are
        not used by the default implementation, but may be used by subclass
        implementations.

        Args:

            json_dict : The dictionary to serialize
        """
        if orjson is not None:
            return orjson.dumps(json_dict, option=orjson.OPT_INDENT_2, **kwargs).decode(
                "utf-8"
            )
        else:
            return json.dumps(json_dict, *args, indent=2, **kwargs)

    def stac_object_from_dict(
        self,
        d: dict[str, Any],
        href: HREF | None = None,
        root: Catalog | None = None,
        preserve_dict: bool = True,
    ) -> STACObject:
        """Deserializes a :class:`~pystac.STACObject` subclass instance from a
        dictionary.

        Args:

            d : The dictionary to deserialize
            href : Optional href to associate with the STAC object
            root : Optional root :class:`~pystac.Catalog` to associate with the
                STAC object.
            preserve_dict: If ``False``, the dict parameter ``d`` may be modified
                during this method call. Otherwise, the dict is not mutated.
                Defaults to ``True``, which results in a deepcopy of the
                parameter. Set to ``False`` when possible to avoid the performance
                hit of a deepcopy.
        """
        href_str = None if href is None else str(os.fspath(href))
        if identify_stac_object_type(d) == pystac.STACObjectType.ITEM:
            collection_cache = None
            if root is not None:
                collection_cache = root._resolved_objects.as_collection_cache()

            # Merge common properties in case this is an older STAC object.
            merge_common_properties(
                d, json_href=href_str, collection_cache=collection_cache
            )

        info = identify_stac_object(d)
        d = migrate_to_latest(d, info)

        if info.object_type == pystac.STACObjectType.CATALOG:
            result = pystac.Catalog.from_dict(
                d, href=href_str, root=root, migrate=True, preserve_dict=preserve_dict
            )
            result._stac_io = self
            return result

        if info.object_type == pystac.STACObjectType.COLLECTION:
            return pystac.Collection.from_dict(
                d, href=href_str, root=root, migrate=True, preserve_dict=preserve_dict
            )

        if info.object_type == pystac.STACObjectType.ITEM:
            return pystac.Item.from_dict(
                d, href=href_str, root=root, migrate=True, preserve_dict=preserve_dict
            )

        raise ValueError(f"Unknown STAC object type {info.object_type}")

    def read_json(self, source: HREF, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Read a dict from the given source.

        See :func:`StacIO.read_text <pystac.StacIO.read_text>` for usage of
        str vs Link as a parameter.

        Args:
            source : The source from which to read.
            *args : Additional positional arguments to be passed to
                :meth:`StacIO.read_text`.
            **kwargs : Additional keyword arguments to be passed to
                :meth:`StacIO.read_text`.

        Returns:
            dict: A dict representation of the JSON contained in the file at the
            given source.
        """
        txt = self.read_text(source, *args, **kwargs)
        return self.json_loads(txt)

    def read_stac_object(
        self,
        source: HREF,
        root: Catalog | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> STACObject:
        """Read a STACObject from a JSON file at the given source.

        See :func:`StacIO.read_text <pystac.StacIO.read_text>` for usage of
        str vs Link as a parameter.

        Args:
            source : The source from which to read.
            root : Optional root of the catalog for this object.
                If provided, the root's resolved object cache can be used to search for
                previously resolved instances of the STAC object.
            *args : Additional positional arguments to be passed to
                :meth:`StacIO.read_json`.
            **kwargs : Additional keyword arguments to be passed to
                :meth:`StacIO.read_json`.

        Returns:
            STACObject: The deserialized STACObject from the serialized JSON
            contained in the file at the given uri.
        """
        d = self.read_json(source, *args, **kwargs)
        return self.stac_object_from_dict(
            d, href=source, root=root, preserve_dict=False
        )

    def save_json(
        self,
        dest: HREF,
        json_dict: dict[str, Any],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Write a dict to the given URI as JSON.

        See :func:`StacIO.write_text <pystac.StacIO.write_text>` for usage of
        str vs Link as a parameter.

        Args:
            dest : The destination file to write the text to.
            json_dict : The JSON dict to write.
            *args : Additional positional arguments to be passed to
                :meth:`StacIO.json_dumps`.
            **kwargs : Additional keyword arguments to be passed to
                :meth:`StacIO.json_dumps`.
        """
        txt = self.json_dumps(json_dict, *args, **kwargs)
        self.write_text(dest, txt)

    @classmethod
    def set_default(cls, stac_io_class: Callable[[], StacIO]) -> None:
        """Set the default StacIO instance to use."""
        cls._default_io = stac_io_class

    @classmethod
    def default(cls) -> StacIO:
        if cls._default_io is None:
            cls._default_io = DefaultStacIO

        return cls._default_io()


class DefaultStacIO(StacIO):
    def read_text(self, source: HREF, *_: Any, **__: Any) -> str:
        """A concrete implementation of :meth:`StacIO.read_text
        <pystac.StacIO.read_text>`. Converts the ``source`` argument to a string (if it
        is not already) and delegates to :meth:`DefaultStacIO.read_text_from_href` for
        opening and reading the file."""
        href = str(os.fspath(source))
        return self.read_text_from_href(href)

    def read_text_from_href(self, href: str) -> str:
        """Reads file as a UTF-8 string.

        If ``href`` has a "scheme" (e.g. if it starts with "https://") then this will
        use :func:`urllib.request.urlopen` (or func:`urllib3.PoolManager().request`
        if available) to open the file and read the contents; otherwise, :func:`open`
        will be used to open a local file.

        Args:

            href : The URI of the file to open.
        """
        href_contents: str
        if _is_url(href):
            import logging
            from urllib.error import HTTPError
            from urllib.request import Request, urlopen

            logger = logging.getLogger(__name__)
            try:
                logger.debug(f"GET {href} Headers: {self.headers}")
                if HAS_URLLIB3:
                    http = urllib3.PoolManager()
                    with http.request(
                        "GET",
                        href,
                        headers={
                            "User-Agent": f"pystac/{pystac.__version__}",
                            **self.headers,
                        },
                        preload_content=False,  # type: ignore
                    ) as f:
                        href_contents = f.read().decode("utf-8")
                else:
                    req = Request(
                        href,
                        headers={
                            "User-Agent": f"pystac/{pystac.__version__}",
                            **self.headers,
                        },
                    )
                    with urlopen(req) as f:
                        href_contents = f.read().decode("utf-8")

            except HTTPError as e:
                raise Exception(f"Could not read uri {href}") from e
        else:
            href = safe_urlparse(href).path
            with open(href, encoding="utf-8") as f:
                href_contents = f.read()
        return href_contents

    def write_text(self, dest: HREF, txt: str, *_: Any, **__: Any) -> None:
        """A concrete implementation of :meth:`StacIO.write_text
        <pystac.StacIO.write_text>`. Converts the ``dest`` argument to a string (if it
        is not already) and delegates to
        W:meth:`~pystac.stac_io.DefaultStacIO.write_text_to_href` for
        opening and reading the file."""
        href = str(os.fspath(dest))
        return self.write_text_to_href(href, txt)

    def write_text_to_href(self, href: str, txt: str) -> None:
        """Writes text to file using UTF-8 encoding.

        This implementation uses :func:`open` and therefore can only write to the local
        file system.

        Args:

            href : The path to which the file will be written.
            txt : The string content to write to the file.
        """
        if _is_url(href):
            raise NotImplementedError("DefaultStacIO cannot write to urls")
        href = safe_urlparse(href).path
        dirname = os.path.dirname(href)
        if dirname != "" and not os.path.isdir(dirname):
            os.makedirs(dirname)
        with open(href, "w", encoding="utf-8") as f:
            f.write(txt)


class DuplicateKeyReportingMixin(StacIO):
    """A mixin for :class:`pystac.StacIO` implementations that will report
    on duplicate keys in the JSON being read in.

    See https://github.com/stac-utils/pystac/issues/313
    """

    def json_loads(self, txt: str, *_: Any, **__: Any) -> dict[str, Any]:
        """Overwrites :meth:`StacIO.json_loads <pystac.StacIO.json_loads>` as the
        internal method used by :class:`DuplicateKeyReportingMixin` for deserializing
        a JSON string to a dictionary while checking for duplicate object keys.

        Raises:

            pystac.DuplicateObjectKeyError : If a duplicate object key is found.
        """
        result: dict[str, Any] = json.loads(
            txt, object_pairs_hook=self._report_duplicate_object_names
        )
        return result

    def read_json(self, source: HREF, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Overwrites :meth:`StacIO.read_json <pystac.StacIO.read_json>` for
        deserializing a JSON file to a dictionary while checking for duplicate object
        keys.

        Raises:

            pystac.DuplicateObjectKeyError : If a duplicate object key is found.
        """
        txt = self.read_text(source, *args, **kwargs)
        try:
            return self.json_loads(txt, source=source)
        except pystac.DuplicateObjectKeyError as e:
            url = str(os.fspath(source))
            msg = str(e) + f" in {url}"
            raise pystac.DuplicateObjectKeyError(msg)

    @staticmethod
    def _report_duplicate_object_names(
        object_pairs: list[tuple[str, Any]],
    ) -> dict[str, Any]:
        result: dict[str, Any] = {}
        for key, value in object_pairs:
            if key in result:
                raise pystac.DuplicateObjectKeyError(
                    f'Found duplicate object name "{key}"'
                )
            else:
                result[key] = value
        return result


if HAS_URLLIB3:
    from typing import cast

    from urllib3 import PoolManager
    from urllib3.util import Retry

    class RetryStacIO(DefaultStacIO):
        """A customized StacIO that retries requests, using
        :py:class:`urllib3.util.retry.Retry`.

        The headers are passed to :py:class:`DefaultStacIO`. If retry is not
        provided, a default retry is used.

        To use this class, you'll need to install PySTAC with urllib3:

        .. code-block:: shell

            pip install pystac[urllib3]

        """

        def __init__(
            self,
            headers: dict[str, str] | None = None,
            retry: Retry | None = None,
        ):
            super().__init__(headers)

            self.retry = retry or Retry()
            """The :py:class:`urllib3.util.retry.Retry` to use with all reading network
            requests."""

        def read_text_from_href(self, href: str) -> str:
            """Reads file as a UTF-8 string, with retry support.

            Args:
                href : The URI of the file to open.
            """
            if _is_url(href):
                from urllib.error import HTTPError

                # TODO provide a pooled StacIO to enable more efficient network
                # access (probably named `PooledStacIO`).
                http = PoolManager()
                try:
                    response = http.request(
                        "GET",
                        href,
                        headers={
                            "User-Agent": f"pystac/{pystac.__version__}",
                            **self.headers,
                        },
                        retries=self.retry,  # type: ignore
                    )
                    return cast(str, response.data.decode("utf-8"))
                except HTTPError as e:
                    raise Exception(f"Could not read uri {href}") from e
            else:
                return super().read_text_from_href(href)



import json
from typing import Any

import pytest

from pystac import (
    Asset,
    Catalog,
    Collection,
    ExtensionNotImplemented,
    ExtensionTypeError,
    Item,
)
from pystac.extensions.file import ByteOrder, FileExtension, MappingObject
from tests.utils import TestCases, assert_to_from_dict

FILE_ITEM_EXAMPLE_URI = TestCases.get_path("data-files/file/item.json")
FILE_COLLECTION_EXAMPLE_URI = TestCases.get_path("data-files/file/collection.json")
FILE_CATALOG_EXAMPLE_URI = TestCases.get_path("data-files/file/catalog.json")


@pytest.fixture
def ext_item() -> Item:
    return Item.from_file(FILE_ITEM_EXAMPLE_URI)


@pytest.fixture
def ext_collection() -> Collection:
    return Collection.from_file(FILE_COLLECTION_EXAMPLE_URI)


@pytest.fixture
def ext_catalog() -> Catalog:
    return Catalog.from_file(FILE_CATALOG_EXAMPLE_URI)


def test_byte_order_to_str() -> None:
    assert ByteOrder.LITTLE_ENDIAN.value == "little-endian"
    assert ByteOrder.BIG_ENDIAN.value == "big-endian"


def test_mapping_object_create() -> None:
    values = [0, 1]
    summary = "clouds"
    m = MappingObject.create(values, summary)

    assert m.values == values
    assert m.summary == summary


def test_mapping_object_set_properties() -> None:
    values = [0, 1]
    summary = "clouds"
    m = MappingObject.create(values, summary)

    new_values = [3, 4]
    new_summary = "cloud shadow"
    m.summary = new_summary
    m.values = new_values

    assert m.values == new_values
    assert m.summary == new_summary


def test_mapping_object_apply() -> None:
    values = [0, 1]
    summary = "clouds"
    m = MappingObject.create(values, summary)

    new_values = [3, 4]
    new_summary = "cloud shadow"
    m.apply(new_values, new_summary)

    assert m.values == new_values
    assert m.summary == new_summary


def test_to_from_dict() -> None:
    with open(FILE_ITEM_EXAMPLE_URI) as f:
        item_dict = json.load(f)
    assert_to_from_dict(Item, item_dict)


@pytest.mark.vcr()
def test_validate_item(ext_item: Item) -> None:
    ext_item.validate()


@pytest.mark.vcr()
def test_validate_collection(ext_collection: Collection) -> None:
    ext_collection.validate()


@pytest.mark.vcr()
def test_validate_catalog(ext_catalog: Catalog) -> None:
    ext_catalog.validate()


@pytest.mark.parametrize(
    "asset_name,field,value",
    [
        ("thumbnail", "size", 146484),
        ("measurement", "header_size", 4096),
        ("thumbnail", "checksum", "90e40210f52acd32b09769d3b1871b420789456c"),
        ("thumbnail", "byte_order", ByteOrder.BIG_ENDIAN),
        (
            "calibrations",
            "local_path",
            (
                "S1A_EW_GRDM_1SSH_20181103T235855_20181103T235955_024430_02AD5D_5616.SAFE"
                "/annotation/calibration/calibration-s1a-ew-grd-hh-20181103t235855-201811"
                "03t235955-024430-02ad5d-001.xml"
            ),
        ),
    ],
)
def test_get_field_on_asset(
    ext_item: Item, asset_name: str, field: str, value: Any
) -> None:
    asset = ext_item.assets[asset_name]

    prop = asset.extra_fields[f"file:{field}"]
    attr = getattr(FileExtension.ext(asset), field)

    assert attr is not None
    assert attr == prop == value


@pytest.mark.vcr()
@pytest.mark.parametrize(
    "asset_name,field,value",
    [
        ("thumbnail", "size", 1),
        ("measurement", "header_size", 8192),
        ("thumbnail", "checksum", "90e40210163700a8a6501eccd00b6d3b44ddaed0"),
        ("thumbnail", "byte_order", ByteOrder.LITTLE_ENDIAN),
        ("calibrations", "local_path", "different-file.xml"),
    ],
)
def test_set_field_on_asset(
    ext_item: Item, asset_name: str, field: str, value: Any
) -> None:
    asset = ext_item.assets[asset_name]

    original = asset.extra_fields[f"file:{field}"]
    setattr(FileExtension.ext(asset), field, value)
    new = asset.extra_fields[f"file:{field}"]

    assert new != original
    assert new == value
    assert ext_item.validate()


@pytest.mark.parametrize(
    "link_rel,field,value",
    [
        ("item", "size", 8675309),
    ],
)
def test_get_field_on_link(
    ext_catalog: Catalog, link_rel: str, field: str, value: Any
) -> None:
    link = ext_catalog.get_links(rel=link_rel)[0]

    prop = link.extra_fields[f"file:{field}"]
    attr = getattr(link.ext.file, field)

    assert attr is not None
    assert attr == prop == value


@pytest.mark.vcr()
@pytest.mark.parametrize(
    "link_rel,field,value",
    [
        ("about", "size", 129302),
        ("about", "header_size", 4092),
        ("about", "checksum", "90e40210163700a8a6501eccd00b6d3b44ddaedb"),
        ("about", "byte_order", ByteOrder.BIG_ENDIAN),
        ("about", "local_path", "a/path"),
    ],
)
def test_set_field_on_link(
    ext_collection: Collection, link_rel: str, field: str, value: Any
) -> None:
    link = ext_collection.get_links(rel=link_rel)[0]

    setattr(link.ext.file, field, value)
    new = link.extra_fields[f"file:{field}"]

    assert new == value
    assert ext_collection.validate()


def test_item_asset_values(ext_item: Item) -> None:
    # Set/get
    asset = ext_item.assets["thumbnail"]
    file_ext = FileExtension.ext(asset)
    mapping_obj = {"values": [0], "summary": "clouds"}

    file_ext.apply(values=[MappingObject.from_dict(mapping_obj)])

    assert asset.extra_fields["file:values"] == [mapping_obj]

    values_field = asset.extra_fields["file:values"]
    assert isinstance(values_field, list)
    assert all(isinstance(mapping_obj, dict) for mapping_obj in values_field)


def test_item_asset_apply(ext_item: Item) -> None:
    asset = ext_item.assets["thumbnail"]

    new_checksum = "90e40210163700a8a6501eccd00b6d3b44ddaed0"
    new_size = 1
    new_header_size = 8192
    new_mapping_obj = {"values": [0], "summary": "clouds"}
    new_values = [MappingObject.from_dict(new_mapping_obj)]
    new_byte_order = ByteOrder.LITTLE_ENDIAN
    new_local_path = "./my/thumbnail.png"

    assert asset.ext.file.checksum != new_checksum
    assert asset.ext.file.size != new_size
    assert asset.ext.file.header_size != new_header_size
    assert asset.ext.file.values != new_values
    assert asset.ext.file.byte_order != new_byte_order
    assert asset.ext.file.local_path != new_local_path

    asset.ext.file.apply(
        byte_order=new_byte_order,
        checksum=new_checksum,
        size=new_size,
        header_size=new_header_size,
        values=new_values,
        local_path=new_local_path,
    )

    assert asset.extra_fields["file:checksum"] == new_checksum
    assert asset.extra_fields["file:size"] == new_size
    assert asset.extra_fields["file:header_size"] == new_header_size
    assert asset.extra_fields["file:values"] == [new_mapping_obj]
    assert asset.extra_fields["file:byte_order"] == new_byte_order
    assert asset.extra_fields["file:local_path"] == new_local_path


def test_repr(ext_item: Item) -> None:
    asset = ext_item.assets["thumbnail"]
    file_ext = FileExtension.ext(asset)

    assert str(file_ext) == f"<AssetFileExtension Asset href={asset.href}>"


def test_migrates_old_checksum() -> None:
    example_path = TestCases.get_path(
        "data-files/examples/1.0.0-beta.2/extensions/checksum/examples/sentinel1.json"
    )
    item = Item.from_file(example_path)

    assert FileExtension.has_extension(item)
    assert (
        FileExtension.ext(item.assets["noises"]).checksum
        == "90e40210a30d1711e81a4b11ef67b28744321659"
    )


def test_extension_not_implemented(sample_item: Item) -> None:
    # Should raise exception if Item does not include extension URI
    asset = sample_item.assets["thumbnail"]

    with pytest.raises(ExtensionNotImplemented):
        FileExtension.ext(asset)

    # Should succeed if Asset has no owner
    ownerless_asset = Asset.from_dict(asset.to_dict())
    FileExtension.ext(ownerless_asset)


def test_ext_add_to(sample_item: Item) -> None:
    asset = sample_item.assets["thumbnail"]
    assert FileExtension.get_schema_uri() not in sample_item.stac_extensions

    FileExtension.ext(asset, add_if_missing=True)
    assert FileExtension.get_schema_uri() in sample_item.stac_extensions


def test_should_raise_exception_when_passing_invalid_extension_object(
    sample_item: Item,
) -> None:
    with pytest.raises(
        ExtensionTypeError, match="FileExtension does not apply to type 'Item'"
    ):
        FileExtension.ext(sample_item, add_if_missing=True)  # type: ignore


@pytest.mark.vcr()
def test_migrate_from_v2_0_0() -> None:
    item = Item.from_file(
        "https://raw.githubusercontent.com/stac-extensions/file/v2.0.0/examples/item.json"
    )
    asset = item.assets["thumbnail"]
    assert FileExtension.get_schema_uri() in item.stac_extensions
    assert asset.ext.file.local_path is None


@pytest.mark.vcr()
def test_migrate_from_v1_0_0() -> None:
    with pytest.warns(
        UserWarning,
        match=(
            r"Assets \['measurement', 'thumbnail'\] contain fields: "
            r"\['file:data_type'\] which were removed"
        ),
    ):
        item = Item.from_file(
            "https://raw.githubusercontent.com/stac-extensions/file/v1.0.0/examples/item.json"
        )
    asset = item.assets["measurement"]
    assert FileExtension.get_schema_uri() in item.stac_extensions

    assert not hasattr(asset.ext.file, "data_type")
    assert asset.extra_fields["file:data_type"] == "uint16"
    assert asset.ext.file.local_path is None



from __future__ import annotations

import re
import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import (
    Any,
    Generic,
    TypeVar,
    cast,
)

import pystac
from pystac.stac_object import S

VERSION_REGEX = re.compile("/v[0-9].[0-9].*/")


class SummariesExtension:
    """Base class for extending the properties in :attr:`pystac.Collection.summaries`
    to include properties defined by a STAC Extension.

    This class should generally not be instantiated directly. Instead, create an
    extension-specific class that inherits from this class and instantiate that. See
    :class:`~pystac.extensions.eo.SummariesEOExtension` for an example."""

    summaries: pystac.Summaries
    """The summaries for the :class:`~pystac.Collection` being extended."""

    def __init__(self, collection: pystac.Collection) -> None:
        self.summaries = collection.summaries

    def _set_summary(
        self,
        prop_key: str,
        v: list[Any] | pystac.RangeSummary[Any] | dict[str, Any] | None,
    ) -> None:
        if v is None:
            self.summaries.remove(prop_key)
        else:
            self.summaries.add(prop_key, v)


P = TypeVar("P")


class PropertiesExtension(ABC):
    """Abstract base class for extending the properties of an :class:`~pystac.Item`
    to include properties defined by a STAC Extension.

    This class should not be instantiated directly. Instead, create an
    extension-specific class that inherits from this class and instantiate that. See
    :class:`~pystac.extensions.eo.EOExtension` for an example.
    """

    properties: dict[str, Any]
    """The properties that this extension wraps.

    The extension which implements PropertiesExtension can use ``_get_property`` and
    ``_set_property`` to get and set values on this instance. Note that _set_properties
    mutates the properties directly."""

    additional_read_properties: Iterable[dict[str, Any]] | None = None
    """Additional read-only properties accessible from the extended object.

    These are used when extending an :class:`~pystac.Asset` to give access to the
    properties of the owning :class:`~pystac.Item`. If a property exists in both
    ``additional_read_properties`` and ``properties``, the value in
    ``additional_read_properties`` will take precedence.
    """

    def _get_property(self, prop_name: str, _typ: type[P]) -> P | None:
        maybe_property: P | None = self.properties.get(prop_name)
        if maybe_property is not None:
            return maybe_property
        if self.additional_read_properties is not None:
            for props in self.additional_read_properties:
                maybe_additional_property: P | None = props.get(prop_name)
                if maybe_additional_property is not None:
                    return maybe_additional_property
        return None

    def _set_property(
        self, prop_name: str, v: Any | None, pop_if_none: bool = True
    ) -> None:
        if v is None and pop_if_none:
            self.properties.pop(prop_name, None)
        elif isinstance(v, list):
            self.properties[prop_name] = [
                x.to_dict() if hasattr(x, "to_dict") else x for x in v
            ]
        else:
            self.properties[prop_name] = v


class ExtensionManagementMixin(Generic[S], ABC):
    """Abstract base class with methods for adding and removing extensions from STAC
    Objects. This class is generic over the type of object being extended (e.g.
    :class:`~pystac.Item`).

    Concrete extension implementations should inherit from this class and either
    provide a concrete type or a bounded type variable.

    See :class:`~pystac.extensions.eo.EOExtension` for an example implementation.
    """

    @classmethod
    @abstractmethod
    def get_schema_uri(cls) -> str:
        """Gets the schema URI associated with this extension."""
        raise NotImplementedError

    @classmethod
    def get_schema_uris(cls) -> list[str]:
        """Gets a list of schema URIs associated with this extension."""
        warnings.warn(
            "get_schema_uris is deprecated and will be removed in v2",
            DeprecationWarning,
        )
        return [cls.get_schema_uri()]

    @classmethod
    def add_to(cls, obj: S) -> None:
        """Add the schema URI for this extension to the
        :attr:`~pystac.STACObject.stac_extensions` list for the given object, if it is
        not already present."""
        if obj.stac_extensions is None:
            obj.stac_extensions = [cls.get_schema_uri()]
        elif not cls.has_extension(obj):
            obj.stac_extensions.append(cls.get_schema_uri())

    @classmethod
    def remove_from(cls, obj: S) -> None:
        """Remove the schema URI for this extension from the
        :attr:`pystac.STACObject.stac_extensions` list for the given object."""
        if obj.stac_extensions is not None:
            obj.stac_extensions = [
                uri for uri in obj.stac_extensions if uri != cls.get_schema_uri()
            ]

    @classmethod
    def has_extension(cls, obj: S) -> bool:
        """Check if the given object implements this extension by checking
        :attr:`pystac.STACObject.stac_extensions` for this extension's schema URI."""
        schema_startswith = VERSION_REGEX.split(cls.get_schema_uri())[0] + "/"

        return obj.stac_extensions is not None and any(
            uri.startswith(schema_startswith) for uri in obj.stac_extensions
        )

    @classmethod
    def validate_owner_has_extension(
        cls,
        asset: pystac.Asset | pystac.ItemAssetDefinition,
        add_if_missing: bool = False,
    ) -> None:
        """
        DEPRECATED

        .. deprecated:: 1.9
            Use :meth:`ensure_owner_has_extension` instead.

        Given an :class:`~pystac.Asset`, checks if the asset's owner has this
        extension's schema URI in its :attr:`~pystac.STACObject.stac_extensions` list.
        If ``add_if_missing`` is ``True``, the schema URI will be added to the owner.

        Args:
            asset : The asset whose owner should be validated.
            add_if_missing : Whether to add the schema URI to the owner if it is
                not already present. Defaults to False.

        Raises:
            STACError : If ``add_if_missing`` is ``True`` and ``asset.owner`` is
                ``None``.
        """
        warnings.warn(
            "ensure_owner_has_extension is deprecated. "
            "Use ensure_owner_has_extension instead",
            DeprecationWarning,
        )
        return cls.ensure_owner_has_extension(asset, add_if_missing)

    @classmethod
    def ensure_owner_has_extension(
        cls,
        asset_or_link: pystac.Asset | pystac.ItemAssetDefinition | pystac.Link,
        add_if_missing: bool = False,
    ) -> None:
        """Given an :class:`~pystac.Asset`, checks if the asset's owner has this
        extension's schema URI in its :attr:`~pystac.STACObject.stac_extensions` list.
        If ``add_if_missing`` is ``True``, the schema URI will be added to the owner.

        Args:
            asset : The asset whose owner should be validated.
            add_if_missing : Whether to add the schema URI to the owner if it is
                not already present. Defaults to False.

        Raises:
            STACError : If ``add_if_missing`` is ``True`` and ``asset.owner`` is
                ``None``.
        """
        if asset_or_link.owner is None:
            if add_if_missing:
                raise pystac.STACError(
                    f"Attempted to use add_if_missing=True for a {type(asset_or_link)} "
                    "with no owner. Use .set_owner or set add_if_missing=False."
                )
            else:
                return
        return cls.ensure_has_extension(cast(S, asset_or_link.owner), add_if_missing)

    @classmethod
    def validate_has_extension(cls, obj: S, add_if_missing: bool = False) -> None:
        """
        DEPRECATED

        .. deprecated:: 1.9
            Use :meth:`ensure_has_extension` instead.

        Given a :class:`~pystac.STACObject`, checks if the object has this
        extension's schema URI in its :attr:`~pystac.STACObject.stac_extensions` list.
        If ``add_if_missing`` is ``True``, the schema URI will be added to the object.

        Args:
            obj : The object to validate.
            add_if_missing : Whether to add the schema URI to the object if it is
                not already present. Defaults to False.
        """
        warnings.warn(
            "validate_has_extension is deprecated. Use ensure_has_extension instead",
            DeprecationWarning,
        )

        return cls.ensure_has_extension(obj, add_if_missing)

    @classmethod
    def ensure_has_extension(cls, obj: S, add_if_missing: bool = False) -> None:
        """Given a :class:`~pystac.STACObject`, checks if the object has this
        extension's schema URI in its :attr:`~pystac.STACObject.stac_extensions` list.
        If ``add_if_missing`` is ``True``, the schema URI will be added to the object.

        Args:
            obj : The object to validate.
            add_if_missing : Whether to add the schema URI to the object if it is
                not already present. Defaults to False.
        """
        if add_if_missing:
            cls.add_to(obj)

        if not cls.has_extension(obj):
            name = getattr(cls, "name", cls.__name__)
            suggestion = (
                f"``obj.ext.add('{name}')"
                if hasattr(cls, "name")
                else f"``{name}.add_to(obj)``"
            )

            raise pystac.ExtensionNotImplemented(
                f"Extension '{name}' is not implemented on object."
                f"STAC producers can add the extension using {suggestion}"
            )

    @classmethod
    def _ext_error_message(cls, obj: Any) -> str:
        contents = [f"{cls.__name__} does not apply to type '{type(obj).__name__}'"]
        if hasattr(cls, "summaries") and isinstance(obj, pystac.Collection):
            hint = f"Hint: Did you mean to use `{cls.__name__}.summaries` instead?"
            contents.append(hint)
        return ". ".join(contents)



"""
Implements the `Item Asset Definition Object
<https://github.com/radiantearth/stac-spec/blob/v1.1.0/collection-spec/collection-spec.md#item-asset-definition-object>`__
for use as values in the :attr:`~pystac.Collection.item_assets` dict.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pystac

if TYPE_CHECKING:
    from pystac.extensions.ext import ItemAssetExt


ASSET_TITLE_PROP = "title"
ASSET_DESC_PROP = "description"
ASSET_TYPE_PROP = "type"
ASSET_ROLES_PROP = "roles"


class ItemAssetDefinition:
    """Implementation of the `Item Asset Definition Object
    <https://github.com/radiantearth/stac-spec/blob/v1.1.0/collection-spec/collection-spec.md#item-asset-definition-object>`__
    for use as values in the :attr:`~pystac.Collection.item_assets` dict.
    """

    properties: dict[str, Any]

    owner: pystac.Collection | None

    def __init__(
        self, properties: dict[str, Any], owner: pystac.Collection | None = None
    ) -> None:
        self.properties = properties
        self.owner = owner

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, ItemAssetDefinition):
            return NotImplemented
        return self.to_dict() == o.to_dict()

    @classmethod
    def create(
        cls,
        title: str | None,
        description: str | None,
        media_type: str | None,
        roles: list[str] | None,
        extra_fields: dict[str, Any] | None = None,
    ) -> ItemAssetDefinition:
        """
        Creates a new asset definition.

        Args:
            title : Displayed title for clients and users.
            description : Description of the Asset providing additional details,
                such as how it was processed or created.
                `CommonMark 0.29 <http://commonmark.org/>`__ syntax MAY be used
                for rich text representation.
            media_type : `media type\
                <https://github.com/radiantearth/stac-spec/tree/v1.1.0/catalog-spec/catalog-spec.md#media-types>`__
                 of the asset.
            roles : `semantic roles
                <https://github.com/radiantearth/stac-spec/tree/v1.1.0/item-spec/item-spec.md#asset-role-types>`__
                of the asset, similar to the use of rel in links.
            extra_fields : Additional fields on the asset definition, e.g. from
                extensions.
        """
        asset_defn = cls({})
        asset_defn.apply(
            title=title,
            description=description,
            media_type=media_type,
            roles=roles,
            extra_fields=extra_fields,
        )
        return asset_defn

    def apply(
        self,
        title: str | None,
        description: str | None,
        media_type: str | None,
        roles: list[str] | None,
        extra_fields: dict[str, Any] | None = None,
    ) -> None:
        """
        Sets the properties for this asset definition.

        Args:
            title : Displayed title for clients and users.
            description : Description of the Asset providing additional details,
                such as how it was processed or created.
                `CommonMark 0.29 <http://commonmark.org/>`__ syntax MAY be used
                for rich text representation.
            media_type : `media type\
                <https://github.com/radiantearth/stac-spec/tree/v1.0.0/catalog-spec/catalog-spec.md#media-types>`__
                 of the asset.
            roles : `semantic roles
                <https://github.com/radiantearth/stac-spec/tree/v1.0.0/item-spec/item-spec.md#asset-role-types>`__
                of the asset, similar to the use of rel in links.
            extra_fields : Additional fields on the asset definition, e.g. from
                extensions.
        """
        if extra_fields:
            self.properties.update(extra_fields)
        self.title = title
        self.description = description
        self.media_type = media_type
        self.roles = roles
        self.owner = None

    def set_owner(self, obj: pystac.Collection) -> None:
        """Sets the owning item of this ItemAssetDefinition.

        The owning item will be used to resolve relative HREFs of this asset.

        Args:
            obj: The Collection that owns this asset.
        """
        self.owner = obj

    @property
    def title(self) -> str | None:
        """Gets or sets the displayed title for clients and users."""
        return self.properties.get(ASSET_TITLE_PROP)

    @title.setter
    def title(self, v: str | None) -> None:
        if v is None:
            self.properties.pop(ASSET_TITLE_PROP, None)
        else:
            self.properties[ASSET_TITLE_PROP] = v

    @property
    def description(self) -> str | None:
        """Gets or sets a description of the Asset providing additional details, such as
        how it was processed or created. `CommonMark 0.29 <http://commonmark.org/>`__
        syntax MAY be used for rich text representation."""
        return self.properties.get(ASSET_DESC_PROP)

    @description.setter
    def description(self, v: str | None) -> None:
        if v is None:
            self.properties.pop(ASSET_DESC_PROP, None)
        else:
            self.properties[ASSET_DESC_PROP] = v

    @property
    def media_type(self) -> str | None:
        """Gets or sets the `media type
        <https://github.com/radiantearth/stac-spec/tree/v1.0.0/catalog-spec/catalog-spec.md#media-types>`__
        of the asset."""
        return self.properties.get(ASSET_TYPE_PROP)

    @media_type.setter
    def media_type(self, v: str | None) -> None:
        if v is None:
            self.properties.pop(ASSET_TYPE_PROP, None)
        else:
            self.properties[ASSET_TYPE_PROP] = v

    @property
    def roles(self) -> list[str] | None:
        """Gets or sets the `semantic roles
        <https://github.com/radiantearth/stac-spec/tree/v1.0.0/item-spec/item-spec.md#asset-role-types>`__
        of the asset, similar to the use of rel in links."""
        return self.properties.get(ASSET_ROLES_PROP)

    @roles.setter
    def roles(self, v: list[str] | None) -> None:
        if v is None:
            self.properties.pop(ASSET_ROLES_PROP, None)
        else:
            self.properties[ASSET_ROLES_PROP] = v

    def to_dict(self) -> dict[str, Any]:
        """Returns a dictionary representing this ``ItemAssetDefinition``."""
        from copy import deepcopy

        return deepcopy(self.properties)

    def create_asset(self, href: str) -> pystac.Asset:
        """Creates a new :class:`~pystac.Asset` instance using the fields from this
        ``ItemAssetDefinition`` and the given ``href``."""
        return pystac.Asset(
            href=href,
            title=self.title,
            description=self.description,
            media_type=self.media_type,
            roles=self.roles,
            extra_fields={
                k: v
                for k, v in self.properties.items()
                if k
                not in {
                    ASSET_TITLE_PROP,
                    ASSET_DESC_PROP,
                    ASSET_TYPE_PROP,
                    ASSET_ROLES_PROP,
                }
            },
        )

    @property
    def ext(self) -> ItemAssetExt:
        """Accessor for extension classes on this item_asset

        Example::

            collection.item_assets["data"].ext.proj.epsg = 4326
        """
        from pystac.extensions.ext import ItemAssetExt

        return ItemAssetExt(stac_object=self)


class _ItemAssets(dict):  # type:ignore
    """Private class for exposing item_assets as a dict

    This class coerces values to ``ItemAssetDefinition``s and
    sets that owner on all ``ItemAssetDefinition``s to the collection
    that it is owned by.
    """

    collection: pystac.Collection

    def __init__(self, collection: pystac.Collection) -> None:
        self.collection = collection
        if not collection.extra_fields.get("item_assets"):
            collection.extra_fields["item_assets"] = {}
        self.update(collection.extra_fields["item_assets"])

    def __setitem__(self, key: str, value: Any) -> None:
        if isinstance(value, ItemAssetDefinition):
            asset_definition = value
            asset_definition.set_owner(self.collection)
        else:
            asset_definition = ItemAssetDefinition(value, self.collection)
        self.collection.extra_fields["item_assets"][key] = asset_definition.properties
        super().__setitem__(key, asset_definition)

    def update(self, *args: Any, **kwargs: Any) -> None:
        for k, v in dict(*args, **kwargs).items():
            self[k] = v



"""Implements the :stac-ext:`Grid Extension <grid>`."""

from __future__ import annotations

import re
from re import Pattern
from typing import Any, Literal

import pystac
from pystac.extensions.base import ExtensionManagementMixin, PropertiesExtension
from pystac.extensions.hooks import ExtensionHooks

SCHEMA_URI: str = "https://stac-extensions.github.io/grid/v1.1.0/schema.json"
SCHEMA_URIS: list[str] = [
    "https://stac-extensions.github.io/grid/v1.0.0/schema.json",
    SCHEMA_URI,
]
PREFIX: str = "grid:"

# Field names
CODE_PROP: str = PREFIX + "code"  # required

CODE_REGEX: str = r"[A-Z0-9]+-[-_.A-Za-z0-9]+"
CODE_PATTERN: Pattern[str] = re.compile(CODE_REGEX)


def validated_code(v: str) -> str:
    if not isinstance(v, str):
        raise ValueError("Invalid Grid code: must be str")
    if not CODE_PATTERN.fullmatch(v):
        raise ValueError(
            f"Invalid Grid code: {v} does not match the regex {CODE_REGEX}"
        )
    return v


class GridExtension(
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """A concrete implementation of :class:`~pystac.extensions.grid.GridExtension`
    on an :class:`~pystac.Item`
    that extends the properties of the Item to include properties defined in the
    :stac-ext:`Grid Extension <grid>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`~pystac.extensions.grid.GridExtension.ext` on an :class:`~pystac.Item`
    to extend it.

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> proj_ext = GridExtension.ext(item)
    """

    name: Literal["grid"] = "grid"

    item: pystac.Item
    """The :class:`~pystac.Item` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Item` properties, including extension properties."""

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return f"<ItemGridExtension Item id={self.item.id}>"

    def apply(self, code: str) -> None:
        """Applies Grid extension properties to the extended Item.

        Args:
            code : REQUIRED. The code of the Item's grid location.
        """
        self.code = validated_code(code)

    @property
    def code(self) -> str | None:
        """Get or sets the grid code of the datasource."""
        return self._get_property(CODE_PROP, str)

    @code.setter
    def code(self, v: str) -> None:
        self._set_property(CODE_PROP, validated_code(v), pop_if_none=False)

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def get_schema_uris(cls) -> list[str]:
        import warnings

        warnings.warn(
            "get_schema_uris is deprecated and will be removed in v2",
            DeprecationWarning,
        )
        return SCHEMA_URIS

    @classmethod
    def ext(cls, obj: pystac.Item, add_if_missing: bool = False) -> GridExtension:
        """Extends the given STAC Object with properties from the :stac-ext:`Grid
        Extension <grid>`.

        This extension can be applied to instances of :class:`~pystac.Item`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return GridExtension(obj)
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))


class GridExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids: set[str] = {*[uri for uri in SCHEMA_URIS if uri != SCHEMA_URI]}
    stac_object_types = {pystac.STACObjectType.ITEM}


GRID_EXTENSION_HOOKS: ExtensionHooks = GridExtensionHooks()



from __future__ import annotations

import os
from collections.abc import Callable, Iterable, Iterator
from copy import deepcopy
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    cast,
)

import pystac
import pystac.media_type
from pystac.cache import ResolvedObjectCache
from pystac.errors import STACError, STACTypeError
from pystac.layout import (
    APILayoutStrategy,
    BestPracticesLayoutStrategy,
    HrefLayoutStrategy,
    LayoutTemplate,
)
from pystac.link import Link
from pystac.serialization import (
    identify_stac_object,
    identify_stac_object_type,
    migrate_to_latest,
)
from pystac.stac_object import STACObject, STACObjectType
from pystac.utils import (
    HREF,
    StringEnum,
    _is_url,
    is_absolute_href,
    make_absolute_href,
    make_relative_href,
)

if TYPE_CHECKING:
    from pystac.asset import Asset
    from pystac.collection import Collection
    from pystac.extensions.ext import CatalogExt
    from pystac.item import Item

#: Generalized version of :class:`Catalog`
C = TypeVar("C", bound="Catalog")


class CatalogType(StringEnum):
    SELF_CONTAINED = "SELF_CONTAINED"
    """A 'self-contained catalog' is one that is designed for portability.
    Users may want to download an online catalog from and be able to use it on their
    local computer, so all links need to be relative.

    See:
        :stac-spec:`The best practices documentation on self-contained catalogs
        <best-practices.md#self-contained-catalogs>`
    """

    ABSOLUTE_PUBLISHED = "ABSOLUTE_PUBLISHED"
    """
    Absolute Published Catalog is a catalog that uses absolute links for everything,
    both in the links objects and in the asset hrefs.

    See:
        :stac-spec:`The best practices documentation on published catalogs
        <best-practices.md#published-catalogs>`
    """

    RELATIVE_PUBLISHED = "RELATIVE_PUBLISHED"
    """
    Relative Published Catalog is a catalog that uses relative links for everything, but
    includes an absolute self link at the root catalog, to identify its online location.

    See:
        :stac-spec:`The best practices documentation on published catalogs
        <best-practices.md#published-catalogs>`
    """

    @classmethod
    def determine_type(cls, stac_json: dict[str, Any]) -> CatalogType | None:
        """Determines the catalog type based on a STAC JSON dict.

        Only applies to Catalogs or Collections

        Args:
            stac_json : The STAC JSON dict to determine the catalog type

        Returns:
            Optional[CatalogType]: The catalog type of the catalog or collection.
            Will return None if it cannot be determined.
        """
        self_link = None
        relative = False
        for link in stac_json["links"]:
            if link["rel"] == pystac.RelType.SELF:
                self_link = link
            else:
                relative |= not is_absolute_href(link["href"])

        if self_link:
            if relative:
                return cls.RELATIVE_PUBLISHED
            else:
                return cls.ABSOLUTE_PUBLISHED
        else:
            if relative:
                return cls.SELF_CONTAINED
            else:
                return None


class Catalog(STACObject):
    """A PySTAC Catalog represents a STAC catalog in memory.

    A Catalog is a :class:`~pystac.STACObject` that may contain children,
    which are instances of :class:`~pystac.Catalog` or :class:`~pystac.Collection`,
    as well as :class:`~pystac.Item` s.

    Args:
        id : Identifier for the catalog. Must be unique within the STAC.
        description : Detailed multi-line description to fully explain the catalog.
            `CommonMark 0.29 syntax <https://commonmark.org/>`_ MAY be used for rich
            text representation.
        title : Optional short descriptive one-line title for the catalog.
        stac_extensions : Optional list of extensions the Catalog implements.
        href : Optional HREF for this catalog, which be set as the
            catalog's self link's HREF.
        catalog_type : Optional catalog type for this catalog. Must
            be one of the values in :class:`~pystac.CatalogType`.
        strategy : The layout strategy to use for setting the
            HREFs of the catalog child objects and items.
            If not provided, it will default to the strategy of the root and fallback to
            :class:`~pystac.layout.BestPracticesLayoutStrategy`.
    """

    catalog_type: CatalogType
    """The catalog type. Defaults to :attr:`CatalogType.ABSOLUTE_PUBLISHED`."""

    description: str
    """Detailed multi-line description to fully explain the catalog."""

    extra_fields: dict[str, Any]
    """Extra fields that are part of the top-level JSON properties of the Catalog."""

    id: str
    """Identifier for the catalog."""

    links: list[Link]
    """A list of :class:`~pystac.Link` objects representing all links associated with
    this Catalog."""

    title: str | None
    """Optional short descriptive one-line title for the catalog."""

    stac_extensions: list[str]
    """List of extensions the Catalog implements."""

    _resolved_objects: ResolvedObjectCache

    STAC_OBJECT_TYPE = pystac.STACObjectType.CATALOG

    _stac_io: pystac.StacIO | None = None
    """Optional instance of StacIO that will be used by default
    for any IO operations on objects contained by this catalog.
    Set while reading in a catalog. This is set when a catalog
    is read by a StacIO instance."""

    DEFAULT_FILE_NAME = "catalog.json"
    """Default file name that will be given to this STAC object in
    a canonical format.
    """

    _fallback_strategy: HrefLayoutStrategy = BestPracticesLayoutStrategy()
    """Fallback layout strategy"""

    def __init__(
        self,
        id: str,
        description: str,
        title: str | None = None,
        stac_extensions: list[str] | None = None,
        extra_fields: dict[str, Any] | None = None,
        href: str | None = None,
        catalog_type: CatalogType = CatalogType.ABSOLUTE_PUBLISHED,
        strategy: HrefLayoutStrategy | None = None,
    ):
        super().__init__(stac_extensions or [])

        self.id = id
        self.description = description
        self.title = title
        if extra_fields is None:
            self.extra_fields = {}
        else:
            self.extra_fields = extra_fields

        self._resolved_objects = ResolvedObjectCache()

        self.add_link(Link.root(self))

        if href is not None:
            self.set_self_href(href)

        self.catalog_type: CatalogType = catalog_type

        self.strategy: HrefLayoutStrategy | None = strategy

        self._resolved_objects.cache(self)

    def __repr__(self) -> str:
        return f"<Catalog id={self.id}>"

    def set_root(self, root: Catalog | None) -> None:
        STACObject.set_root(self, root)
        if root is not None:
            root._resolved_objects = ResolvedObjectCache.merge(
                root._resolved_objects, self._resolved_objects
            )

            # Walk through resolved object links and update the root
            for link in self.links:
                if link.rel == pystac.RelType.CHILD or link.rel == pystac.RelType.ITEM:
                    target = link.target
                    if isinstance(target, STACObject):
                        target.set_root(root)

    def is_relative(self) -> bool:
        return self.catalog_type in [
            CatalogType.RELATIVE_PUBLISHED,
            CatalogType.SELF_CONTAINED,
        ]

    def _get_strategy(self, strategy: HrefLayoutStrategy | None) -> HrefLayoutStrategy:
        if strategy is not None:
            return strategy
        elif self.strategy is not None:
            return self.strategy
        elif root := self.get_root():
            if root.strategy is not None:
                return root.strategy
            else:
                return root._fallback_strategy
        else:
            return self._fallback_strategy

    def add_child(
        self,
        child: Catalog | Collection,
        title: str | None = None,
        strategy: HrefLayoutStrategy | None = None,
        set_parent: bool = True,
    ) -> Link:
        """Adds a link to a child :class:`~pystac.Catalog` or
        :class:`~pystac.Collection`.

        This method will set the child's parent to this object and potentially
        override its self_link (unless ``set_parent`` is False).

        It will always set its root to this Catalog's root.

        Args:
            child : The child to add.
            title : Optional title to give to the :class:`~pystac.Link`
            strategy : The layout strategy to use for setting the
                self href of the child. If not provided, defaults to
                the layout strategy of the parent or root and falls back to
                :class:`~pystac.layout.BestPracticesLayoutStrategy`.
            set_parent : Whether to set the parent on the child as well.
                Defaults to True.

        Returns:
            Link: The link created for the child
        """

        # Prevent typo confusion
        if isinstance(child, pystac.Item):
            raise pystac.STACError("Cannot add item as child. Use add_item instead.")

        strategy = self._get_strategy(strategy)

        child.set_root(self.get_root())
        if set_parent:
            child.set_parent(self)
        else:
            child._allow_parent_to_override_href = False

        # set self link
        self_href = self.get_self_href()
        if self_href and set_parent:
            child_href = strategy.get_href(child, self_href)
            child.set_self_href(child_href)

        child_link = Link.child(child, title=title)
        self.add_link(child_link)
        return child_link

    def add_children(
        self,
        children: Iterable[Catalog | Collection],
        strategy: HrefLayoutStrategy | None = None,
    ) -> list[Link]:
        """Adds links to multiple :class:`~pystac.Catalog` or `~pystac.Collection`
        objects. This method will set each child's parent to this object, and their
        root to this Catalog's root.

        Args:
            children : The children to add.
            strategy : The layout strategy to use for setting the
                self href of the children. If not provided, defaults to
                the layout strategy of the parent or root and falls back to
                :class:`~pystac.layout.BestPracticesLayoutStrategy`.

        Returns:
            List[Link]: An array of links created for the children
        """
        return [self.add_child(child, strategy=strategy) for child in children]

    def add_item(
        self,
        item: Item,
        title: str | None = None,
        strategy: HrefLayoutStrategy | None = None,
        set_parent: bool = True,
    ) -> Link:
        """Adds a link to an :class:`~pystac.Item`.

        This method will set the item's parent to this object and potentially
        override its self_link (unless ``set_parent`` is False)

        It will always set its root to this Catalog's root.

        Args:
            item : The item to add.
            title : Optional title to give to the :class:`~pystac.Link`
            strategy : The layout strategy to use for setting the
                self href of the item. If not provided, defaults to
                the layout strategy of the parent or root and falls back to
                :class:`~pystac.layout.BestPracticesLayoutStrategy`.
            set_parent : Whether to set the parent on the item as well.
                Defaults to True.

        Returns:
            Link: The link created for the item
        """

        # Prevent typo confusion
        if isinstance(item, pystac.Catalog):
            raise pystac.STACError("Cannot add catalog as item. Use add_child instead.")

        strategy = self._get_strategy(strategy)

        item.set_root(self.get_root())
        if set_parent:
            item.set_parent(self)
        else:
            item._allow_parent_to_override_href = False

        # set self link
        self_href = self.get_self_href()
        if self_href and set_parent:
            item_href = strategy.get_href(item, self_href)
            item.set_self_href(item_href)

        item_link = Link.item(item, title=title)
        self.add_link(item_link)
        return item_link

    def add_items(
        self,
        items: Iterable[Item],
        strategy: HrefLayoutStrategy | None = None,
    ) -> list[Link]:
        """Adds links to multiple :class:`Items <pystac.Item>`.

        This method will set each item's parent to this object, and their root to
        this Catalog's root.

        Args:
            items : The items to add.
            strategy : The layout strategy to use for setting the
                self href of the items. If not provided, defaults to
                the layout strategy of the parent or root and falls back to
                :class:`~pystac.layout.BestPracticesLayoutStrategy`.

        Returns:
            List[Link]: A list of links created for the item
        """
        return [self.add_item(item, strategy=strategy) for item in items]

    def get_child(
        self, id: str, recursive: bool = False, sort_links_by_id: bool = True
    ) -> Catalog | Collection | None:
        """Gets the child of this catalog with the given ID, if it exists.

        Args:
            id : The ID of the child to find.
            recursive : If True, search this catalog and all children for the
                item; otherwise, only search the children of this catalog. Defaults
                to False.
            sort_links_by_id : If True, links containing the ID will be checked
                first. If links do not contain the ID then setting this to False
                will improve performance. Defaults to True.

        Return:
            Catalog or Collection or None: The child with the given ID,
            or None if not found.
        """
        if not recursive:
            children: Iterable[pystac.Catalog | pystac.Collection]
            if not sort_links_by_id:
                children = self.get_children()
            else:

                def sort_function(links: list[Link]) -> list[Link]:
                    return sorted(
                        links,
                        key=lambda x: (href := x.get_href()) is None or id not in href,
                    )

                children = map(
                    lambda x: cast(pystac.Catalog | pystac.Collection, x),
                    self.get_stac_objects(
                        pystac.RelType.CHILD, modify_links=sort_function
                    ),
                )
            return next((c for c in children if c.id == id), None)
        else:
            for root, _, _ in self.walk():
                child = root.get_child(id, recursive=False)
                if child is not None:
                    return child
            return None

    def get_children(self) -> Iterable[Catalog | Collection]:
        """Return all children of this catalog.

        Return:
            Iterable[Catalog or Collection]: Iterable of children whose parent
            is this catalog.
        """
        return map(
            lambda x: cast(pystac.Catalog | pystac.Collection, x),
            self.get_stac_objects(pystac.RelType.CHILD),
        )

    def get_collections(self) -> Iterable[Collection]:
        """Return all children of this catalog that are :class:`~pystac.Collection`
        instances."""
        return map(
            lambda x: cast(pystac.Collection, x),
            self.get_stac_objects(pystac.RelType.CHILD, pystac.Collection),
        )

    def get_all_collections(self) -> Iterable[Collection]:
        """Get all collections from this catalog and all subcatalogs. Will traverse
        any subcatalogs recursively."""
        yield from self.get_collections()
        for child in self.get_children():
            yield from child.get_all_collections()

    def get_child_links(self) -> list[Link]:
        """Return all child links of this catalog.

        Return:
            List[Link]: List of links of this catalog with ``rel == 'child'``
        """
        return self.get_links(
            rel=pystac.RelType.CHILD,
            media_type=pystac.media_type.STAC_JSON,
        )

    def clear_children(self) -> None:
        """Removes all children from this catalog.

        Return:
            Catalog: Returns ``self``
        """
        child_ids = [child.id for child in self.get_children()]
        for child_id in child_ids:
            self.remove_child(child_id)

    def remove_child(self, child_id: str) -> None:
        """Removes a child from this catalog.

        Args:
            child_id : The ID of the child to remove.
        """
        new_links: list[pystac.Link] = []
        root = self.get_root()
        for link in self.links:
            if link.rel != pystac.RelType.CHILD:
                new_links.append(link)
            else:
                link.resolve_stac_object(root=root)
                child = cast("Catalog", link.target)
                if child.id != child_id:
                    new_links.append(link)
                else:
                    child.set_parent(None)
                    child.set_root(None)
        self.links = new_links

    def get_item(self, id: str, recursive: bool = False) -> Item | None:
        """
        DEPRECATED.

        .. deprecated:: 1.8
            Use ``next(pystac.Catalog.get_items(id), None)`` instead.

        Returns an item with a given ID.

        Args:
            id : The ID of the item to find.
            recursive : If True, search this catalog and all children for the
                item; otherwise, only search the items of this catalog. Defaults
                to False.

        Return:
            Item or None: The item with the given ID, or None if not found.
        """
        import warnings

        warnings.warn(
            "get_item is deprecated and will be removed in v2. "
            "Use next(self.get_items(id), None) instead",
            DeprecationWarning,
        )
        if not recursive:
            return next((i for i in self.get_items() if i.id == id), None)
        else:
            for root, _, _ in self.walk():
                item = root.get_item(id, recursive=False)
                if item is not None:
                    return item
            return None

    def get_items(self, *ids: str, recursive: bool = False) -> Iterator[Item]:
        """Return all items or specific items of this catalog.

        Args:
            *ids : The IDs of the items to include.
            recursive : If True, search this catalog and all children for the
                item; otherwise, only search the items of this catalog. Defaults
                to False.

        Return:
            Iterator[Item]: Generator of items whose parent is this catalog, and
                (if recursive) all catalogs or collections connected to this catalog
                through child links.
        """
        from itertools import chain

        items: Iterator[Item]
        if not recursive:
            items = map(
                lambda x: cast(pystac.Item, x),
                self.get_stac_objects(pystac.RelType.ITEM),
            )
        else:
            items = chain(
                self.get_items(recursive=False),
                *(child.get_items(recursive=True) for child in self.get_children()),
            )
        if ids:
            yield from (i for i in items if i.id in ids)
        else:
            yield from items

    def clear_items(self) -> None:
        """Removes all items from this catalog.

        Return:
            Catalog: Returns ``self``
        """
        for link in self.get_item_links():
            if link.is_resolved():
                item = cast(pystac.Item, link.target)
                item.set_parent(None)
                item.set_root(None)

        self.links = [link for link in self.links if link.rel != pystac.RelType.ITEM]

    def remove_item(self, item_id: str) -> None:
        """Removes an item from this catalog.

        Args:
            item_id : The ID of the item to remove.
        """
        new_links: list[pystac.Link] = []
        root = self.get_root()
        for link in self.links:
            if link.rel != pystac.RelType.ITEM:
                new_links.append(link)
            else:
                link.resolve_stac_object(root=root)
                item = cast(pystac.Item, link.target)
                if item.id != item_id:
                    new_links.append(link)
                else:
                    item.set_parent(None)
                    item.set_root(None)
        self.links = new_links

    def get_all_items(self) -> Iterator[Item]:
        """
        DEPRECATED.

        .. deprecated:: 1.8
            Use ``pystac.Catalog.get_items(recursive=True)`` instead.

        Get all items from this catalog and all subcatalogs. Will traverse
        any subcatalogs recursively.

        Returns:
            Generator[Item]: All items that belong to this catalog, and all
                catalogs or collections connected to this catalog through
                child links.
        """
        import warnings
        from itertools import chain

        warnings.warn(
            "get_all_items is deprecated and will be removed in v2",
            DeprecationWarning,
        )
        return chain(
            self.get_items(),
            *(child.get_items(recursive=True) for child in self.get_children()),
        )

    def get_item_links(self) -> list[Link]:
        """Return all item links of this catalog.

        Return:
            List[Link]: List of links of this catalog with ``rel == 'item'``
        """
        return self.get_links(
            rel=pystac.RelType.ITEM, media_type=pystac.media_type.STAC_JSON
        )

    def to_dict(
        self, include_self_link: bool = True, transform_hrefs: bool = True
    ) -> dict[str, Any]:
        links = [
            x
            for x in self.links
            if x.rel != pystac.RelType.ROOT or x.get_href(transform_hrefs) is not None
        ]
        if not include_self_link:
            links = [x for x in links if x.rel != pystac.RelType.SELF]

        d: dict[str, Any] = {
            "type": self.STAC_OBJECT_TYPE.value.title(),
            "id": self.id,
            "stac_version": pystac.get_stac_version(),
            "description": self.description,
            "links": [link.to_dict(transform_href=transform_hrefs) for link in links],
        }

        if self.stac_extensions:
            d["stac_extensions"] = self.stac_extensions

        for key in self.extra_fields:
            d[key] = self.extra_fields[key]

        if self.title is not None:
            d["title"] = self.title

        return d

    def clone(self) -> Catalog:
        cls = self.__class__
        clone = cls(
            id=self.id,
            description=self.description,
            title=self.title,
            stac_extensions=self.stac_extensions.copy(),
            extra_fields=deepcopy(self.extra_fields),
            catalog_type=self.catalog_type,
        )
        clone._resolved_objects.cache(clone)

        for link in self.links:
            if link.rel == pystac.RelType.ROOT:
                # Catalog __init__ sets correct root to clone; don't reset
                # if the root link points to self
                root_is_self = link.is_resolved() and link.target is self
                if not root_is_self:
                    clone.set_root(None)
                    clone.add_link(link.clone())
            else:
                clone.add_link(link.clone())

        return clone

    def make_all_asset_hrefs_relative(self) -> None:
        """Recursively makes all the HREFs of assets in this catalog relative"""
        for item in self.get_items(recursive=True):
            item.make_asset_hrefs_relative()
        for collection in self.get_all_collections():
            collection.make_asset_hrefs_relative()

    def make_all_asset_hrefs_absolute(self) -> None:
        """Recursively makes all the HREFs of assets in this catalog absolute"""
        for item in self.get_items(recursive=True):
            item.make_asset_hrefs_absolute()
        for collection in self.get_all_collections():
            collection.make_asset_hrefs_absolute()

    def normalize_and_save(
        self,
        root_href: str,
        catalog_type: CatalogType | None = None,
        strategy: HrefLayoutStrategy | None = None,
        stac_io: pystac.StacIO | None = None,
        skip_unresolved: bool = False,
    ) -> None:
        """Normalizes link HREFs to the given root_href, and saves the catalog.

        This is a convenience method that simply calls :func:`Catalog.normalize_hrefs
        <pystac.Catalog.normalize_hrefs>` and :func:`Catalog.save <pystac.Catalog.save>`
        in sequence.

        Args:
            root_href : The absolute HREF that all links will be normalized
                against.
            catalog_type : The catalog type that dictates the structure of
                the catalog to save. Use a member of :class:`~pystac.CatalogType`.
                Defaults to the root catalog.catalog_type or the current catalog
                catalog_type if there is no root catalog.
            strategy : The layout strategy to use in setting the
                HREFS for this catalog. If not provided, defaults to
                the layout strategy of the parent or root and falls back to
                :class:`~pystac.layout.BestPracticesLayoutStrategy`
            stac_io : Optional instance of :class:`~pystac.StacIO` to use. If not
                provided, will use the instance set while reading in the catalog,
                or the default instance if this is not available.
            skip_unresolved : Skip unresolved links when normalizing the tree.
                Defaults to False. Because unresolved links are not saved, this
                argument can be used to normalize and save only newly-added
                objects.
        """
        self.normalize_hrefs(
            root_href, strategy=strategy, skip_unresolved=skip_unresolved
        )
        self.save(catalog_type, stac_io=stac_io)

    def normalize_hrefs(
        self,
        root_href: str,
        strategy: HrefLayoutStrategy | None = None,
        skip_unresolved: bool = False,
    ) -> None:
        """Normalize HREFs will regenerate all link HREFs based on
        an absolute root_href and the canonical catalog layout as specified
        in the STAC specification's best practices.

        This method mutates the entire catalog tree, unless ``skip_unresolved``
        is True, in which case only resolved links are modified. This is useful
        in the case when you have loaded a large catalog and you've added a few
        items/children, and you only want to update those newly-added objects,
        not the whole tree.

        Args:
            root_href : The absolute HREF that all links will be normalized against.
            strategy : The layout strategy to use in setting the HREFS
                for this catalog. If not provided, defaults to
                the layout strategy of the parent or root and falls back to
                :class:`~pystac.layout.BestPracticesLayoutStrategy`
            skip_unresolved : Skip unresolved links when normalizing the tree.
                Defaults to False.

        See:
            :stac-spec:`STAC best practices document <best-practices.md#catalog-layout>`
            for the canonical layout of a STAC.
        """

        _strategy = self._get_strategy(strategy)

        # Normalizing requires an absolute path
        if not is_absolute_href(root_href):
            root_href = make_absolute_href(root_href, os.getcwd(), start_is_dir=True)

        if isinstance(_strategy, APILayoutStrategy) and not _is_url(root_href):
            raise STACError("When using APILayoutStrategy the root_href must be a URL")

        def process_item(
            item: Item, _root_href: str, is_root: bool, parent: Catalog | None
        ) -> Callable[[], None] | None:
            if not skip_unresolved:
                item.resolve_links()

            # Abort as the intended parent is not the actual parent
            # https://github.com/stac-utils/pystac/issues/1116
            if parent is not None and item.get_parent() != parent:
                return None

            new_self_href = _strategy.get_href(item, _root_href, is_root)

            def fn() -> None:
                item.set_self_href(new_self_href)

            return fn

        def process_catalog(
            cat: Catalog,
            _root_href: str,
            is_root: bool,
            parent: Catalog | None = None,
        ) -> list[Callable[[], None]]:
            setter_funcs: list[Callable[[], None]] = []

            if not skip_unresolved:
                cat.resolve_links()

            # Abort as the intended parent is not the actual parent
            # https://github.com/stac-utils/pystac/issues/1116
            if parent is not None and cat.get_parent() != parent:
                return setter_funcs

            new_self_href = _strategy.get_href(cat, _root_href, is_root)
            new_root = new_self_href

            for link in cat.get_links():
                if skip_unresolved and not link.is_resolved():
                    continue
                elif link.rel == pystac.RelType.ITEM:
                    link.resolve_stac_object(root=self.get_root())
                    item_fn = process_item(
                        cast(pystac.Item, link.target), new_root, is_root, cat
                    )
                    if item_fn is not None:
                        setter_funcs.append(item_fn)
                elif link.rel == pystac.RelType.CHILD:
                    link.resolve_stac_object(root=self.get_root())
                    setter_funcs.extend(
                        process_catalog(
                            cast(pystac.Catalog | pystac.Collection, link.target),
                            new_root,
                            is_root=False,
                            parent=cat,
                        )
                    )

            def fn() -> None:
                cat.set_self_href(new_self_href)

            setter_funcs.append(fn)

            return setter_funcs

        # Collect functions that will actually mutate the objects.
        # Delay mutation as setting hrefs while walking the catalog
        # can result in bad links.
        setter_funcs = process_catalog(self, root_href, is_root=True)

        for fn in setter_funcs:
            fn()

    def generate_subcatalogs(
        self,
        template: str,
        defaults: dict[str, Any] | None = None,
        parent_ids: list[str] | None = None,
    ) -> list[Catalog]:
        """Walks through the catalog and generates subcatalogs
        for items based on the template string.

        See :class:`~pystac.layout.LayoutTemplate`
        for details on the construction of template strings. This template string
        will be applied to the items, and subcatalogs will be created that separate
        and organize the items based on template values.

        Args:
            template :   A template string that
                can be consumed by a :class:`~pystac.layout.LayoutTemplate`
            defaults :  Default values for the template variables
                that will be used if the property cannot be found on
                the item.
            parent_ids : Optional list of the parent catalogs'
                identifiers. If the bottom-most subcatalogs already match the
                template, no subcatalog is added.

        Returns:
            list[Catalog]: List of new catalogs created
        """
        result: list[Catalog] = []
        parent_ids = parent_ids or list()
        parent_ids.append(self.id)
        for child in self.get_children():
            result.extend(
                child.generate_subcatalogs(
                    template, defaults=defaults, parent_ids=parent_ids.copy()
                )
            )

        layout_template = LayoutTemplate(template, defaults=defaults)

        keep_item_links: list[Link] = []
        item_links = [lk for lk in self.links if lk.rel == pystac.RelType.ITEM]
        for link in item_links:
            link.resolve_stac_object(root=self.get_root())
            item = cast(pystac.Item, link.target)
            subcat_ids = layout_template.substitute(item).split("/")
            id_iter = reversed(parent_ids)
            if all([f"{id}" == next(id_iter, None) for id in reversed(subcat_ids)]):
                # Skip items for which the sub-catalog structure already
                # matches the template. The list of parent IDs can include more
                # elements on the root side, so compare the reversed sequences.
                keep_item_links.append(link)
                continue
            curr_parent = self
            for subcat_id in subcat_ids:
                subcat = curr_parent.get_child(subcat_id)
                if subcat is None:
                    subcat_desc = "Catalog of items from {} with id {}".format(
                        curr_parent.id, subcat_id
                    )
                    subcat = pystac.Catalog(id=subcat_id, description=subcat_desc)
                    curr_parent.add_child(subcat)
                    result.append(subcat)
                curr_parent = subcat

            # resolve collection link so when added back points to correct location
            col_link = item.get_single_link(pystac.RelType.COLLECTION)
            if col_link is not None:
                col_link.resolve_stac_object()

            curr_parent.add_item(item)

        # keep only non-item links and item links that have not been moved elsewhere
        self.links = [
            lk for lk in self.links if lk.rel != pystac.RelType.ITEM
        ] + keep_item_links

        return result

    def save(
        self,
        catalog_type: CatalogType | None = None,
        dest_href: str | None = None,
        stac_io: pystac.StacIO | None = None,
    ) -> None:
        """Save this catalog and all it's children/item to files determined by the
        object's self link HREF or a specified path.

        Args:
            catalog_type : The catalog type that dictates the structure of
                the catalog to save. Use a member of :class:`~pystac.CatalogType`.
                If not supplied, the catalog_type of this catalog will be used.
                If that attribute is not set, an exception will be raised.
            dest_href : The location where the catalog is to be saved.
                If not supplied, the catalog's self link HREF is used to determine
                the location of the catalog file and children's files.
            stac_io : Optional instance of :class:`~pystac.StacIO` to use. If not
                provided, will use the instance set while reading in the catalog,
                or the default instance if this is not available.
        Note:
            If the catalog type is ``CatalogType.ABSOLUTE_PUBLISHED``,
            all self links will be included, and hierarchical links be absolute URLs.
            If the catalog type is ``CatalogType.RELATIVE_PUBLISHED``, this catalog's
            self link will be included, but no child catalog will have self links, and
            hierarchical links will be relative URLs
            If the catalog  type is ``CatalogType.SELF_CONTAINED``, no self links will
            be included and hierarchical links will be relative URLs.
        """

        root = self.get_root()
        if root is None:
            raise Exception("There is no root catalog")

        if catalog_type is not None:
            root.catalog_type = catalog_type

        items_include_self_link = root.catalog_type in [CatalogType.ABSOLUTE_PUBLISHED]

        for child_link in self.get_child_links():
            if child_link.is_resolved():
                child = cast(Catalog, child_link.target)
                if dest_href is not None:
                    rel_href = make_relative_href(child.self_href, self.self_href)
                    child_dest_href = make_absolute_href(
                        rel_href, dest_href, start_is_dir=True
                    )
                    child.save(
                        dest_href=os.path.dirname(child_dest_href),
                        stac_io=stac_io,
                    )
                else:
                    child.save(stac_io=stac_io)

        for item_link in self.get_item_links():
            if item_link.is_resolved():
                item = cast(pystac.Item, item_link.target)
                if dest_href is not None:
                    rel_href = make_relative_href(item.self_href, self.self_href)
                    item_dest_href = make_absolute_href(
                        rel_href, dest_href, start_is_dir=True
                    )
                    item.save_object(
                        include_self_link=items_include_self_link,
                        dest_href=item_dest_href,
                        stac_io=stac_io,
                    )
                else:
                    item.save_object(
                        include_self_link=items_include_self_link, stac_io=stac_io
                    )

        include_self_link = False
        # include a self link if this is the root catalog
        # or if ABSOLUTE_PUBLISHED catalog
        if root.catalog_type == CatalogType.ABSOLUTE_PUBLISHED:
            include_self_link = True
        elif root.catalog_type != CatalogType.SELF_CONTAINED:
            root_link = self.get_root_link()
            if root_link and root_link.get_absolute_href() == self.get_self_href():
                include_self_link = True

        catalog_dest_href = None
        if dest_href is not None:
            rel_href = make_relative_href(self.self_href, self.self_href)
            catalog_dest_href = make_absolute_href(
                rel_href, dest_href, start_is_dir=True
            )
        self.save_object(
            include_self_link=include_self_link,
            dest_href=catalog_dest_href,
            stac_io=stac_io,
        )
        if catalog_type is not None:
            self.catalog_type = catalog_type

    def walk(
        self,
    ) -> Iterable[tuple[Catalog, Iterable[Catalog], Iterable[Item]]]:
        """Walks through children and items of catalogs.

        For each catalog in the STAC's tree rooted at this catalog (including this
        catalog itself), it yields a 3-tuple (root, subcatalogs, items). The root in
        that 3-tuple refers to the current catalog being walked, the subcatalogs are any
        catalogs or collections for which the root is a parent, and items represents
        any items that have the root as a parent.

        This has similar functionality to Python's :func:`os.walk`.

        Returns:
           Generator[(Catalog, Generator[Catalog], Generator[Item])]: A generator that
           yields a 3-tuple (parent_catalog, children, items).
        """
        children = self.get_children()
        items = self.get_items()

        yield self, children, items
        for child in self.get_children():
            yield from child.walk()

    def fully_resolve(self) -> None:
        """Resolves every link in this catalog.

        Useful if, e.g., you'd like to read a catalog from a filesystem, upgrade
        every object in the catalog to the latest STAC version, and save it back
        to the filesystem. By default, :py:meth:`~pystac.Catalog.save` skips
        unresolved links.
        """
        for _, _, items in self.walk():
            # items is a generator, so we need to consume it to resolve the
            # items
            for item in items:
                pass

    def validate_all(self, max_items: int | None = None, recursive: bool = True) -> int:
        """Validates each catalog, collection, item contained within this catalog.

        Walks through the children and items of the catalog and validates each
        stac object.

        Args:
            max_items : The maximum number of STAC items to validate. Default
                is None which means, validate them all.
            recursive : Whether to validate catalog, collections, and items contained
                within child objects.

        Returns:
            int : Number of STAC items validated.

        Raises:
            STACValidationError: Raises this error on any item that is invalid.
                Will raise on the first invalid stac object encountered.
        """
        n = 0
        self.validate()
        for child in self.get_children():
            if recursive:
                inner_max_items = None if max_items is None else max_items - n
                n += child.validate_all(max_items=inner_max_items, recursive=True)
            else:
                child.validate()
        for item in self.get_items():
            if max_items is not None and n >= max_items:
                break
            item.validate()
            n += 1
        return n

    def _object_links(self) -> list[str | pystac.RelType]:
        return [
            pystac.RelType.CHILD,
            pystac.RelType.ITEM,
            *pystac.EXTENSION_HOOKS.get_extended_object_links(self),
        ]

    def map_items(
        self,
        item_mapper: Callable[[Item], Item | list[Item]],
    ) -> Catalog:
        """Creates a copy of a catalog, with each item passed through the
        item_mapper function.

        Args:
            item_mapper :   A function that takes in an item, and returns
                either an item or list of items. The item that is passed into the
                item_mapper is a copy, so the method can mutate it safely.

        Returns:
            Catalog: A full copy of this catalog, with items manipulated according
            to the item_mapper function.
        """

        new_cat = self.full_copy()

        def process_catalog(catalog: Catalog) -> None:
            for child in catalog.get_children():
                process_catalog(child)

            item_links: list[Link] = []
            for item_link in catalog.get_item_links():
                item_link.resolve_stac_object(root=self.get_root())
                mapped = item_mapper(cast(pystac.Item, item_link.target))
                if mapped is None:
                    raise Exception("item_mapper cannot return None.")
                if isinstance(mapped, pystac.Item):
                    item_link.target = mapped
                    item_links.append(item_link)
                else:
                    for i in mapped:
                        new_link = item_link.clone()
                        new_link.target = i
                        item_links.append(new_link)
            catalog.clear_items()
            catalog.add_links(item_links)

        process_catalog(new_cat)
        return new_cat

    def map_assets(
        self,
        asset_mapper: Callable[
            [str, Asset],
            Asset | tuple[str, Asset] | dict[str, Asset],
        ],
    ) -> Catalog:
        """Creates a copy of a catalog, with each Asset for each Item passed
        through the asset_mapper function.

        Args:
            asset_mapper : A function that takes in a key and an Asset, and
                returns either an Asset, a (key, Asset), or a dictionary of Assets with
                unique keys. The Asset that is passed into the item_mapper is a copy,
                so the method can mutate it safely.

        Returns:
            Catalog: A full copy of this catalog, with assets manipulated according
            to the asset_mapper function.
        """

        def apply_asset_mapper(
            tup: tuple[str, Asset],
        ) -> list[tuple[str, pystac.Asset]]:
            k, v = tup
            result = asset_mapper(k, v)
            if result is None:
                raise Exception("asset_mapper cannot return None.")
            if isinstance(result, pystac.Asset):
                return [(k, result)]
            elif isinstance(result, tuple):
                return [result]
            else:
                assets = list(result.items())
                if len(assets) < 1:
                    raise Exception("asset_mapper must return a non-empty list")
                return assets

        def item_mapper(item: pystac.Item) -> pystac.Item:
            new_assets = [
                x
                for result in map(apply_asset_mapper, item.assets.items())
                for x in result
            ]
            item.assets = dict(new_assets)
            return item

        return self.map_items(item_mapper)

    def describe(self, include_hrefs: bool = False, _indent: int = 0) -> None:
        """Prints out information about this Catalog and all contained
        STACObjects.

        Args:
            include_hrefs (bool) - If True, print out each object's self link
                HREF along with the object ID.
        """
        s = "{}* {}".format(" " * _indent, self)
        if include_hrefs:
            s += f" {self.get_self_href()}"
        print(s)
        for child in self.get_children():
            child.describe(include_hrefs=include_hrefs, _indent=_indent + 4)
        for item in self.get_items():
            s = "{}* {}".format(" " * (_indent + 2), item)
            if include_hrefs:
                s += f" {item.get_self_href()}"
            print(s)

    @classmethod
    def from_dict(
        cls: type[C],
        d: dict[str, Any],
        href: str | None = None,
        root: Catalog | None = None,
        migrate: bool = True,
        preserve_dict: bool = True,
    ) -> C:
        if migrate:
            info = identify_stac_object(d)
            d = migrate_to_latest(d, info)

        if not cls.matches_object_type(d):
            raise STACTypeError(d, cls)

        catalog_type = CatalogType.determine_type(d)

        if preserve_dict:
            d = deepcopy(d)

        id = d.pop("id")
        description = d.pop("description")
        title = d.pop("title", None)
        stac_extensions = d.pop("stac_extensions", None)
        links = d.pop("links")

        d.pop("stac_version")

        cat = cls(
            id=id,
            description=description,
            title=title,
            stac_extensions=stac_extensions,
            extra_fields=d,
            href=href,
            catalog_type=catalog_type or CatalogType.ABSOLUTE_PUBLISHED,
        )

        for link in links:
            if link["rel"] == pystac.RelType.ROOT:
                # Remove the link that's generated in Catalog's constructor.
                cat.remove_links(pystac.RelType.ROOT)

            if link["rel"] != pystac.RelType.SELF or href is None:
                cat.add_link(Link.from_dict(link))

        if root:
            cat.set_root(root)

        return cat

    def full_copy(
        self, root: Catalog | None = None, parent: Catalog | None = None
    ) -> Catalog:
        return cast(Catalog, super().full_copy(root, parent))

    @classmethod
    def from_file(cls: type[C], href: HREF, stac_io: pystac.StacIO | None = None) -> C:
        if stac_io is None:
            stac_io = pystac.StacIO.default()

        result = super().from_file(href, stac_io)
        result._stac_io = stac_io

        return result

    @classmethod
    def matches_object_type(cls, d: dict[str, Any]) -> bool:
        return identify_stac_object_type(d) == STACObjectType.CATALOG

    @property
    def ext(self) -> CatalogExt:
        """Accessor for extension classes on this catalog

        Example::

            print(collection.ext.version)
        """
        from pystac.extensions.ext import CatalogExt

        return CatalogExt(stac_object=self)
